{%- extends "basic/layout.html" %}
{%- block sidebarlogo %}
  {{ super() }}
  {%- if theme_is_a_pocoo_project %}
    <p class="logo">
      <a href="http://pocoo.org">
        <img
          class="logo"
          src="{{ pathto('_static/pocoo.png', 1) }}"
          alt="A Pocoo Project" />
      </a>
    </p>
  {% endif %}
{% endblock %}
{%- block relbar1 %}{% endblock %}
{%- block relbar2 %}
  {% if theme_github_fork %}
    <a href="http://github.com/{{ theme_github_fork }}">
      <img
        style="position: absolute; top: 0; right: 0; border: 0;"
        src="https://assets2.github.com/img/71eeaab9d563c2b3c590319b398dd35683265e85?repo=&url=http%3A%2F%2Fs3.amazonaws.com%2Fgithub%2Fribbons%2Fforkme_right_gray_6d6d6d.png&path="
        alt="Fork me on GitHub">
    </a>
  {% endif %}
{% endblock %}
