# coding: utf-8
"""
    brownie
    ~~~~~~~

    Brownie provides various different utilities and datastructures, which are
    commonly used in Python applications.

    :copyright: 2010 by Daniel Neuhäuser
    :license: BSD, see LICENSE.rst for details
"""
#: A string containing the full version number.
__version__ = '0.2.2'
#: A tuple containing the full version number, which should be used for
#: comparisons.
__version_info__ = (0, 2, 2)
