# coding: utf-8
"""
    theme_style
    ~~~~~~~~~~~

    :copyright: 2010 by Daniel Neuhäuser
    :license: BSD
"""
from pygments.styles.colorful import ColorfulStyle


class Minimalism(ColorfulStyle):
    background_color = '#eee'
